@@ -3,17 +3,73 @@ from docutils import nodes
 from docutils.statemachine import ViewList
 from sphinx.domains import cpp, c, python
 
+try:
+    from sphinxcontrib import phpdomain as php
+except ImportError:
+    php = None
+
 import re
 import six
 import textwrap
 
 
+class WithContext(object):
+    def __init__(self, parent, context):
+        self.context = context
+        self.parent = parent
+        self.previous = None
+
+    def __enter__(self):
+        assert self.previous is None
+        self.previous = self.parent.context
+        self.parent.set_context(self.context)
+        return self
+
+    def __exit__(self, et, ev, bt):
+        self.parent.context = self.previous
+        self.previous = None
+
+
+class DoxyCPPClassObject(cpp.CPPClassObject):
+    __bases = []
+
+    def parse_definition(self, parser):
+        # add the base classes
+        ast = parser.parse_declaration("class")
+
+        bases = []
+
+        for base in self.__bases:
+            namestr = base.content_[0].value
+
+            # build a name object
+            # TODO: work out if we can use base.refid in a pending_xref somewhere
+            try:
+                parser = cpp.DefinitionParser(namestr, self, self.env.config)
+            except TypeError:
+                # sphinx < 1.5
+                parser = cpp.DefinitionParser(namestr, self)
+            name = parser._parse_nested_name()
+            parser.assert_end()
+
+            bases.append(
+                cpp.ASTBaseClass(name, base.prot, base.virt == 'virtual', False)
+            )
+
+        ast.declaration.bases = bases
+
+        return ast
+
+    def augment(self, bases):
+        self.__bases = bases or []
+
+
 class DomainDirectiveFactory(object):
     # A mapping from node kinds to cpp domain classes and directive names.
     cpp_classes = {
-        'class': (cpp.CPPClassObject, 'class'),
-        'struct': (cpp.CPPClassObject, 'class'),
-        'interface': (cpp.CPPClassObject, 'interface'),
+        'class': (DoxyCPPClassObject, 'class'),
+        'struct': (DoxyCPPClassObject, 'class'),
+        'interface': (DoxyCPPClassObject, 'interface'),
         'function': (cpp.CPPFunctionObject, 'function'),
         'friend': (cpp.CPPFunctionObject, 'function'),
         'slot': (cpp.CPPFunctionObject, 'function'),
@@ -22,9 +78,7 @@ class DomainDirectiveFactory(object):
         'using': (cpp.CPPTypeObject, 'type'),
         'union': (cpp.CPPTypeObject, 'type'),
         'namespace': (cpp.CPPTypeObject, 'type'),
-        # Use CPPClassObject for enum values as the cpp domain doesn't have a directive for
-        # enum values and CPPMemberObject requires a type.
-        'enumvalue': (cpp.CPPClassObject, 'member'),
+        'enumvalue': (cpp.CPPEnumeratorObject, 'enumerator'),
         'define': (c.CObject, 'macro')
     }
 
@@ -33,6 +87,15 @@ class DomainDirectiveFactory(object):
         'variable': (python.PyClassmember, 'attribute')
     }
 
+    if php is not None:
+        php_classes = {
+            'function': (php.PhpNamespacelevel, 'function'),
+            'class': (php.PhpClasslike, 'class'),
+            'attr': (php.PhpClassmember, 'attr'),
+            'method': (php.PhpClassmember, 'method'),
+            'global': (php.PhpGloballevel, 'global'),
+        }
+
     @staticmethod
     def fix_python_signature(sig):
         def_ = 'def '
@@ -49,6 +112,19 @@ class DomainDirectiveFactory(object):
             cls, name = DomainDirectiveFactory.python_classes.get(
                 args[0], (python.PyClasslike, 'class'))
             args[1] = [DomainDirectiveFactory.fix_python_signature(n) for n in args[1]]
+        elif php is not None and domain == 'php':
+            separators = php.separators
+            arg_0 = args[0]
+            if any([separators['method'] in n for n in args[1]]):
+                if any([separators['attr'] in n for n in args[1]]):
+                    arg_0 = 'attr'
+                else:
+                    arg_0 = 'method'
+            else:
+                if arg_0 in ['variable']:
+                    arg_0 = 'global'
+            cls, name = DomainDirectiveFactory.php_classes.get(
+                arg_0, (php.PhpClasslike, 'class'))
         else:
             cls, name = DomainDirectiveFactory.cpp_classes.get(
                 args[0], (cpp.CPPMemberObject, 'member'))
@@ -106,7 +182,7 @@ def get_param_decl(param):
     if not param_name:
         param_decl = param_type
     else:
-        param_decl, number_of_subs = re.subn(r'(\([*&]+)(\))', r'\1' + param_name + r'\2',
+        param_decl, number_of_subs = re.subn(r'(\([*&]+)(\))', r'\g<1>' + param_name + r'\g<2>',
                                              param_type)
         if number_of_subs == 0:
             param_decl = param_type + ' ' + param_name
@@ -248,7 +324,7 @@ class SphinxRenderer(object):
 
         return '::'.join(names)
 
-    def create_template_node(self, decl):
+    def insert_template_node(self, signature, decl):
         """Creates a node for the ``template <...>`` part of the declaration."""
         if not decl.templateparamlist:
             return None
@@ -256,14 +332,19 @@ class SphinxRenderer(object):
         nodes = [self.node_factory.desc_annotation(template, template), self.node_factory.Text('<')]
         nodes.extend(self.render(decl.templateparamlist))
         nodes.append(self.node_factory.Text(">"))
-        signode = self.node_factory.desc_signature()
-        signode.extend(nodes)
-        return signode
 
-    def run_domain_directive(self, kind, names):
+        signode = type(signature)()
+        signode.extend(nodes)
+
+        signature.parent.insert(0, signode)
+
+    def run_domain_directive(self, kind, names, augment=None):
         domain_directive = DomainDirectiveFactory.create(
             self.context.domain, [kind, names] + self.context.directive_args[2:])
 
+        if hasattr(domain_directive, 'augment') and augment is not None:
+            domain_directive.augment(**augment)
+
         # Translate Breathe's no-link option into the standard noindex option.
         if 'no-link' in self.context.directive_args[2]:
             domain_directive.options['noindex'] = True
@@ -350,11 +431,16 @@ class SphinxRenderer(object):
         # We counter that second issue slightly by allowing through single white spaces
         #
         if node.strip():
-            if "<linebreak>" not in node:
-                return [self.node_factory.Text(node)]
-            # Render lines as paragraphs because RST doesn't have line breaks.
-            return [self.node_factory.paragraph('', '', self.node_factory.Text(line))
-                    for line in node.split("<linebreak>")]
+            delimiter = None
+            if "<linebreak>" in node:
+                delimiter = "<linebreak>"
+            elif "\n" in node:
+                delimiter = "\n"
+            if delimiter:
+                # Render lines as paragraphs because RST doesn't have line breaks.
+                return [self.node_factory.paragraph('', '', self.node_factory.Text(line))
+                        for line in node.split(delimiter) if line.strip()]
+            return [self.node_factory.Text(node)]
         if node == six.u(" "):
             return [self.node_factory.Text(node)]
         return []
@@ -377,7 +463,9 @@ class SphinxRenderer(object):
         # Read in the corresponding xml file and process
         file_data = self.compound_parser.parse(node.refid)
 
-        rendered_data = self.render(file_data)
+        parent_context = self.context.create_child_context(file_data)
+        new_context = parent_context.create_child_context(file_data.compounddef)
+        rendered_data = self.render(file_data, parent_context)
 
         if not rendered_data and not render_empty_node:
             return []
@@ -389,7 +477,13 @@ class SphinxRenderer(object):
         def render_signature(file_data, doxygen_target, name, kind):
             # Defer to domains specific directive.
             self.context.directive_args[1] = [self.get_fully_qualified_name()]
-            nodes = self.run_domain_directive(kind, self.context.directive_args[1])
+
+            if kind in ('class', 'struct'):
+                augment = dict(bases=file_data.compounddef.basecompoundref)
+            else:
+                augment = None
+
+            nodes = self.run_domain_directive(kind, self.context.directive_args[1], augment=augment)
             rst_node = nodes[1]
 
             finder = NodeFinder(rst_node.document)
@@ -400,19 +494,19 @@ class SphinxRenderer(object):
             finder.declarator[0] = self.node_factory.desc_annotation(kind + ' ', kind + ' ')
 
             # Check if there is template information and format it as desired
-            template_signode = self.create_template_node(file_data.compounddef)
-            if template_signode:
-                rst_node.insert(0, template_signode)
+            self.insert_template_node(finder.declarator, file_data.compounddef)
             rst_node.children[0].insert(0, doxygen_target)
             return nodes, finder.content
 
         refid = self.get_refid(node.refid)
         render_sig = kwargs.get('render_signature', render_signature)
-        nodes, contentnode = render_sig(file_data, self.target_handler.create_target(refid),
-                                        name, kind)
+        with WithContext(self, new_context):
+            # Pretend that the signature is being rendered in context of the
+            # definition, for proper domain detection
+            nodes, contentnode = render_sig(
+                    file_data, self.target_handler.create_target(refid),
+                    name, kind)
 
-        parent_context = self.context.create_child_context(file_data)
-        new_context = parent_context.create_child_context(file_data.compounddef)
         if file_data.compounddef.includes:
             for include in file_data.compounddef.includes:
                 contentnode.extend(self.render(include, new_context.create_child_context(include)))
@@ -454,18 +548,18 @@ class SphinxRenderer(object):
         ("public-attrib", "Public Members"),
         ("public-slot", "Public Slots"),
         ("signal", "Signal"),
-        ("dcop-func",  "DCOP Function"),
-        ("property",  "Property"),
-        ("event",  "Event"),
+        ("dcop-func", "DCOP Function"),
+        ("property", "Property"),
+        ("event", "Event"),
         ("public-static-func", "Public Static Functions"),
         ("public-static-attrib", "Public Static Attributes"),
-        ("protected-type",  "Protected Types"),
-        ("protected-func",  "Protected Functions"),
-        ("protected-attrib",  "Protected Attributes"),
-        ("protected-slot",  "Protected Slots"),
-        ("protected-static-func",  "Protected Static Functions"),
-        ("protected-static-attrib",  "Protected Static Attributes"),
-        ("package-type",  "Package Types"),
+        ("protected-type", "Protected Types"),
+        ("protected-func", "Protected Functions"),
+        ("protected-attrib", "Protected Attributes"),
+        ("protected-slot", "Protected Slots"),
+        ("protected-static-func", "Protected Static Functions"),
+        ("protected-static-attrib", "Protected Static Attributes"),
+        ("package-type", "Package Types"),
         ("package-func", "Package Functions"),
         ("package-attrib", "Package Attributes"),
         ("package-static-func", "Package Static Functions"),
@@ -473,17 +567,17 @@ class SphinxRenderer(object):
         ("private-type", "Private Types"),
         ("private-func", "Private Functions"),
         ("private-attrib", "Private Members"),
-        ("private-slot",  "Private Slots"),
+        ("private-slot", "Private Slots"),
         ("private-static-func", "Private Static Functions"),
-        ("private-static-attrib",  "Private Static Attributes"),
-        ("friend",  "Friends"),
-        ("related",  "Related"),
-        ("define",  "Defines"),
-        ("prototype",  "Prototypes"),
-        ("typedef",  "Typedefs"),
-        ("enum",  "Enums"),
-        ("func",  "Functions"),
-        ("var",  "Variables"),
+        ("private-static-attrib", "Private Static Attributes"),
+        ("friend", "Friends"),
+        ("related", "Related"),
+        ("define", "Defines"),
+        ("prototype", "Prototypes"),
+        ("typedef", "Typedefs"),
+        ("enum", "Enums"),
+        ("func", "Functions"),
+        ("var", "Variables"),
     ]
 
     def visit_compounddef(self, node):
@@ -507,7 +601,6 @@ class SphinxRenderer(object):
                     *intersperse(output, self.node_factory.Text(', '))
                 )
             )
-        render_list(node.basecompoundref, 'Inherits from ')
         render_list(node.derivedcompoundref, 'Subclassed by ')
 
         section_nodelists = {}
@@ -766,7 +859,7 @@ class SphinxRenderer(object):
 
             lines = node.text.splitlines()
             # Replace the first * on each line with a blank space
-            lines = [text.replace("*", " ", 1) for text in lines]
+            lines = map(lambda text: text.replace("*", " ", 1), lines)
             node.text = "\n".join(lines)
 
         # do we need to strip leading ///?
@@ -774,11 +867,11 @@ class SphinxRenderer(object):
 
             lines = node.text.splitlines()
             # Replace the /// on each line with three blank spaces
-            lines = [text.replace("///", "   ", 1) for text in lines]
+            lines = map(lambda text: text.replace("///", "   ", 1), lines)
             node.text = "\n".join(lines)
 
         # Remove the first line which is "embed:rst[:leading-asterisk]"
-        text = "\n".join(node.text.split("\n")[1:])
+        text = "\n".join(node.text.split(u"\n")[1:])
 
         # Remove starting whitespace
         text = textwrap.dedent(text)
@@ -799,7 +892,7 @@ class SphinxRenderer(object):
 
     def visit_inc(self, node):
 
-        if node.local == "yes":
+        if node.local == u"yes":
             text = '#include "%s"' % node.content_[0].getValue()
         else:
             text = '#include <%s>' % node.content_[0].getValue()
@@ -859,17 +952,18 @@ class SphinxRenderer(object):
 
         refid = self.get_refid(node.refid)
 
-        nodelist = [
-            self.node_factory.pending_xref(
-                "",
-                reftype="ref",
-                refdomain="std",
-                refexplicit=True,
-                refid=refid,
-                reftarget=refid,
-                *nodelist
-            )
-        ]
+        if refid is not None:
+            nodelist = [
+                self.node_factory.pending_xref(
+                    "",
+                    reftype="ref",
+                    refdomain="std",
+                    refexplicit=True,
+                    refid=refid,
+                    reftarget=refid,
+                    *nodelist
+                )
+            ]
 
         return nodelist
 
@@ -901,6 +995,11 @@ class SphinxRenderer(object):
         if node.volatile == 'yes' or node.argsstring.endswith('volatile'):
             signature += ' volatile'
 
+        if node.refqual == 'lvalue':
+            signature += '&'
+        elif node.refqual == 'rvalue':
+            signature += '&&'
+
         self.context.directive_args[1] = [signature]
 
         nodes = self.run_domain_directive(node.kind, self.context.directive_args[1])
@@ -918,9 +1017,7 @@ class SphinxRenderer(object):
 
         finder.content.extend(self.description(node))
 
-        template_node = self.create_template_node(node)
-        if template_node:
-            rst_node.insert(0, template_node)
+        self.insert_template_node(finder.declarator, node)
         return nodes
 
     def visit_define(self, node):
@@ -1192,17 +1289,17 @@ class SphinxRenderer(object):
     }
 
     def render(self, node, context=None):
-        saved_context = self.context
-        self.set_context(context if context else self.context.create_child_context(node))
-        result = []
-        if not self.filter_.allow(self.context.node_stack):
-            pass
-        elif isinstance(node, six.string_types):
-            result = self.visit_unicode(node)
-        else:
-            method = SphinxRenderer.methods.get(node.node_type, SphinxRenderer.visit_unknown)
-            result = method(self, node)
-        self.context = saved_context
+        if context is None:
+            context = self.context.create_child_context(node)
+        with WithContext(self, context):
+            result = []
+            if not self.filter_.allow(self.context.node_stack):
+                pass
+            elif isinstance(node, six.string_types):
+                result = self.visit_unicode(node)
+            else:
+                method = SphinxRenderer.methods.get(node.node_type, SphinxRenderer.visit_unknown)
+                result = method(self, node)
         return result
 
     def render_optional(self, node):
